unit LinkedLabel;

interface
uses
  Messages, Classes, Controls, StdCtrls;

type
  TLinkedLabel = class(TLabel)
  private
    FAssociate:  TControl;	// The associate control
    FCapsLock:   boolean;     // Puts FAssociate into all caps mode
    FGap:        integer;	// The distance between the label and the associate
    FOnTop:      boolean;	// True when the label is on top of the associate
    FOldWinProc: TWndMethod;	// Saves the original value of FAssociate.WindowProc
    FUpdating:   boolean;     // Used to prevent infinite update loops
  protected
    procedure Adjust(MoveLabel: boolean);
    procedure SetGap(Value: integer);
    procedure SetOnTop(Value: boolean);
    procedure SetAssociate(Value: TControl);
    procedure NewWinProc(var Message: TMessage);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure WndProc(var Message: TMessage); override;
  public
    constructor Create(AOwner :TComponent); override;
    destructor Destroy; override;
  published
    property Associate: TControl read FAssociate write SetAssociate;
    property CapsLock:  boolean  read FCapsLock  write FCapsLock;
    property Gap:       integer  read FGap       write SetGap default 8;
    property OnTop:     boolean  read FOnTop     write SetOnTop;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Articles', [TLinkedLabel]);
end;

constructor TLinkedLabel.Create(AOwner: TComponent);
begin
  inherited;
  FGap := 8;
end;

procedure TLinkedLabel.Adjust(MoveLabel: boolean);
var
  dx, dy: integer;

begin
  if(assigned(FAssociate)) then
  begin
    if(FOnTop) then
    begin
      dx := 0;
      dy := Height + FGap;
    end
    else
    begin
      dx := Width + FGap;
      dy := (Height - FAssociate.Height) div 2;
    end;
    if(MoveLabel) then
    begin
      Left := FAssociate.Left - dx;
      Top  := FAssociate.Top - dy;
    end
    else
    begin
      FAssociate.Left := Left + dx;
      FAssociate.Top  := Top + dy;
    end;
  end;
end;

procedure TLinkedLabel.SetGap(Value: integer);
begin
  if(FGap <> Value) then
  begin
    FGap := Value;
    Adjust(TRUE);
  end;
end;

procedure TLinkedLabel.SetOnTop(Value: boolean);
begin
  if(FOnTop <> Value) then
  begin
    FOnTop := Value;
    Adjust(TRUE);
  end;
end;

procedure TLinkedLabel.SetAssociate(Value: TControl);
begin
  if(Value <> FAssociate) then
  begin
    if(assigned(FAssociate)) then
      FAssociate.WindowProc := FOldWinProc;

    FAssociate := Value;

    if(assigned(Value)) then
    begin
      Adjust(TRUE);
      Enabled := FAssociate.Enabled;
      Visible := FAssociate.Visible;
      FOldWinProc := FAssociate.WindowProc;
      FAssociate.WindowProc := NewWinProc;
    end;
  end;
end;

destructor TLinkedLabel.Destroy;
begin
  SetAssociate(nil);
  inherited;
end;

procedure TLinkedLabel.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if((Operation = opRemove) and (AComponent = FAssociate)) then SetAssociate(nil);
end;

procedure TLinkedLabel.NewWinProc(var Message: TMessage);
var
  Ch: char;

begin
  if(assigned(FAssociate) and (not FUpdating)) then
  begin
    FUpdating := TRUE;
    try
      case(Message.Msg) of
        WM_CHAR: if(FCapsLock) then
                 begin
                   Ch := Char(TWMKey(Message).CharCode);
                   if(Ch >= 'a') and (Ch <= 'z') then
                     TWMKey(Message).CharCode := ord(UpCase(Ch));
                 end;
        CM_ENABLEDCHANGED:   Enabled := FAssociate.Enabled;
        CM_VISIBLECHANGED:   Visible := FAssociate.Visible;
        WM_SIZE, WM_MOVE,
        WM_WINDOWPOSCHANGED: Adjust(TRUE);
      end;
    finally
      FUpdating := FALSE;
    end;
  end;
  FOldWinProc(Message);
end;

procedure TLinkedLabel.WndProc(var Message: TMessage);
begin
  if(assigned(FAssociate) and (not FUpdating)) then
  begin
    FUpdating := TRUE;
    try
      case(Message.Msg) of
        CM_ENABLEDCHANGED:   FAssociate.Enabled := Enabled;
        CM_VISIBLECHANGED:   FAssociate.Visible := Visible;
        WM_WINDOWPOSCHANGED: Adjust(FALSE);
      end;
   finally
      FUpdating := FALSE;
    end;
  end;
  inherited;
end;

end.

