unit oaActor;

{ Assistant Package v1.0 }
{ Copyright  1999 Kambiz R. Khojasteh }
{ email: Khojasteh@www.dci.co.ir }

interface

uses
  Windows, Messages, Classes, Controls, Dialogs, Forms, SysUtils, oaTypes;

type

  EActorError = class(Exception);

  TAssistantActor = class(TPersistent)
  private
    FInstance: THandle;
    FLibraryName: String;
    FActorInfo: PActorInfo;
    FOnChange: TNotifyEvent;
    ActionCount: array[TActorAction] of Word;
    procedure LoadActorLibrary(LibName: String);
    procedure FreeActorLibrary;
    function LoadAnimations: Boolean;
    procedure SetLibraryName(Value: String);
    function GetAnimation(Index: Word): TAnimation;
    function GetAnimationInfo(Index: Word): PAnimationInfo;
    function GetActionAnimationCount(Index: TActorAction): Word;
  public
    constructor Create;
    destructor Destroy; override;
    function Active: Boolean;
    function Version: String;
    function Copyright: String;
    function ActorName: String;
    function AnimationSize: TSize;
    function AnimationCount: Word;
    function IndexOf(Action: TActorAction; Level: Word = 0): Word; overload;
    function IndexOf(Name: String): Word; overload;
    property Instance: THandle read FInstance;
    property Animation[Index: Word]: TAnimation read GetAnimation;
    property AnimationInfo[Index: Word]: PAnimationInfo read GetAnimationInfo;
    property ActionAnimationCount[Index: TActorAction]: Word read GetActionAnimationCount;
    property LibraryName: String read FLibraryName write SetLibraryName;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

implementation

const
  ActorLibErrors: array[1..2] of String = (
    '%s'#13#10 + 'File is not a valid actor library',
    '%s'#13#10 + 'Actor library file not found');

constructor TAssistantActor.Create;
begin
  inherited Create;
  FLibraryName := EmptyStr;
  FActorInfo := nil;
  FInstance := 0;
  Randomize;
end;

destructor TAssistantActor.Destroy;
begin
  FreeActorLibrary;
  inherited Destroy;
end;

procedure TAssistantActor.LoadActorLibrary(LibName: String);
var
  OldInstance: THandle;
  OldActor: PActorInfo;
begin
  OldInstance := FInstance;
  OldActor := FActorInfo;
  if FileExists(LibName) then
  begin
    FInstance := LoadLibrary(PChar(LibName));
    if (FInstance <> 0) and not LoadAnimations then
    begin
      FreeLibrary(FInstance);
      FInstance := 0;
    end;
    if FInstance = 0 then
    begin
      FActorInfo := OldActor;
      FInstance := OldInstance;
      raise EActorError.CreateFmt(ActorLibErrors[1], [LibName]);
    end;
  end
  else
    raise EActorError.CreateFmt(ActorLibErrors[2], [LibName]);
end;

procedure TAssistantActor.FreeActorLibrary;
begin
  if FInstance <> 0 then
  begin
    FActorInfo := nil;
    FreeLibrary(FInstance);
    FInstance := 0;
  end;
end;

function TAssistantActor.Active: boolean;
begin
  Result := FInstance <> 0;
end;

procedure TAssistantActor.SetLibraryName(Value: String);
begin
  if (CompareText(LibraryName, Value) <> 0) then
  begin
    if Value <> EmptyStr then
      LoadActorLibrary(Value)
    else
      FreeActorLibrary;
    FLibraryName := Value;
    if Assigned(FOnChange) then
      FOnChange(Self);
  end;
end;

function TAssistantActor.LoadAnimations: Boolean;
var
  GetActorInfo: function: Pointer; stdcall;
  Action: TActorAction;
  ValidLib: Boolean;
  I: Word;
begin
  Result := False;
  if FInstance <> 0 then
  begin
    @GetActorInfo := GetProcAddress(FInstance, 'GetActorInfo');
    if Assigned(GetActorInfo) then
    begin
      try
        FActorInfo := PActorInfo(GetActorInfo);
        ValidLib := Assigned(FActorInfo) and (FActorInfo^.ID = 110) and
                   (FActorInfo^.AnimationCount > 0);
      except
        Exit;
      end;
      if ValidLib then
      begin
        for Action := Low(TActorAction) to High(TActorAction) do
          ActionCount[Action] := 0;
        for I := 1 to FActorInfo^.AnimationCount do
          for Action := Low(TActorAction) to High(TActorAction) do
            if Action in FActorInfo^.Animations[I].Actions then
              Inc(ActionCount[Action]);
        Result := True;
      end;
    end;
  end;
end;

function TAssistantActor.Version: String;
begin
  Result := EmptyStr;
  if Assigned(FActorInfo) and (FActorInfo^.Version <> 0) then
    Result := Format('%d.%2.2d', [FActorInfo^.Version div 100, FActorInfo^.Version mod 100]);
end;

function TAssistantActor.Copyright: String;
begin
  Result := EmptyStr;
  if Assigned(FActorInfo) then
    Result := FActorInfo^.Copyright;
end;

function TAssistantActor.ActorName: String;
begin
  Result := EmptyStr;
  if Assigned(FActorInfo) then
    Result := FActorInfo^.ActorName;
end;

function TAssistantActor.AnimationCount: Word;
begin
  Result := 0;
  if Assigned(FActorInfo) then
    Result := FActorInfo^.AnimationCount;
end;

function TAssistantActor.AnimationSize: TSize;
begin
  Result.CX := 0;
  Result.CY := 0;
  if Assigned(FActorInfo) then
  begin
    Result.CX := FActorInfo^.AnimationWidth;
    Result.CY := FActorInfo^.AnimationHeight;
  end;
end;

function TAssistantActor.GetActionAnimationCount(Index: TActorAction): Word;
begin
  Result := 0;
  if Assigned(FActorInfo) and (Index in [Low(TActorAction)..High(TActorAction)]) then
    Result := ActionCount[Index];
end;

function TAssistantActor.GetAnimationInfo(Index: Word): PAnimationInfo;
begin
  Result := nil;
  if Assigned(FActorInfo) and (Index >= 1) and (Index <= FActorInfo^.AnimationCount) then
    Result := @(FActorInfo^.Animations[Index]);
end;

function TAssistantActor.GetAnimation(Index: Word): TAnimation;
var
  RetIndex: Word;
begin
  FillChar(Result, SizeOf(TAnimation), 0);
  if Assigned(FActorInfo) and (Index >= 1) and (Index <= FActorInfo^.AnimationCount) then
  begin
    Result.Name := FActorInfo^.Animations[Index].Name;
    Result.Actions := FActorInfo^.Animations[Index].Actions;
    Result.HasSound := FActorInfo^.Animations[Index].HasSound;
    Result.ExitType := FActorInfo^.Animations[Index].ExitType;
    if Result.ExitType = etAnimation then
      RetIndex := FActorInfo^.Animations[Index].ReturnIndex
    else
      RetIndex := 0;
    if (RetIndex >= 1) and (RetIndex <= FActorInfo^.AnimationCount) then
      Result.ReturnName := FActorInfo^.Animations[RetIndex].Name
    else
      Result.ReturnName := EmptyStr;
  end;
end;

function TAssistantActor.IndexOf(Action: TActorAction; Level: Word): Word;
var
  FoundLevel, I: Word;
begin
  Result := 0;
  if Assigned(FActorInfo) and (Action <> aaNone) and (ActionCount[Action] > 0) then
  begin
    if Level = 0 then
      Level := Random(ActionCount[Action]) + 1
    else if Level > ActionCount[Action] then
      Level := Level mod ActionCount[Action];
    FoundLevel := 0;
    for I := 1 to FActorInfo^.AnimationCount do
      if Action in FActorInfo^.Animations[I].Actions then
      begin
       Inc(FoundLevel);
       Result := I;
       if FoundLevel >= Level then Break;
      end;
  end;
end;

function TAssistantActor.IndexOf(Name: String): Word;
var
  I: Word;
begin
  Result := 0;
  if Assigned(FActorInfo) and (Name <> EmptyStr) then
    for I := 1 to FActorInfo^.AnimationCount do
      if CompareText(FActorInfo^.Animations[I].Name, Name) = 0 then
      begin
        Result := I;
        Break;
      end;
end;

end.
