unit oaDirect;

{ Office 97 Assistant Package v1.0 }
{ Copyright  1999 Kambiz R. Khojasteh }
{ email: Khojasteh@www.dci.co.ir }

interface

uses
  Windows, Classes, Graphics, ExtCtrls, oaEngine, oaActor, oaTypes;

type

  TRequestQueue = class;

  TAssistantDirector = class(TPersistent)
  private
    FActor: TAssistantActor;
    FAnimator: TAssistantEngine;
    FAutoIdle: Boolean;
    FIdleLevel: TIdleLevel;
    FEnabled: Boolean;
    Queue: TRequestQueue;
    IdleTimer: TTimer;
    CurAnim: PAnimationInfo;
    Returning: Boolean;
    ID: TRequestID;
    IdleID: TRequestID;
    IdlingTime: Boolean;
    IdleLevelTry: Integer;
    FOnPlayStart: TPlayEvent;
    FOnPlayStop: TPlayEvent;
    FOnIdle: TIdleEvent;
    FOnActorChange: TNotifyEvent;
    procedure SetEnabled(Value: Boolean);
    procedure SetAutoIdle(Value: Boolean);
    procedure SetIdleLevel(Value: TIdleLevel);
    procedure ActorChanged(Sender: TObject);
    procedure AnimationStopped(Sender: TObject);
    procedure IdleTimerExpire(Sender: TObject);
    procedure GoToIdleState;
  protected
    procedure PlayQueuedRequest;
  public
    constructor Create;
    destructor Destroy; override;
    function Play(Index: Word): TRequestID; overload;
    function Play(Name: String): TRequestID; overload;
    function Play(Action: TActorAction): TRequestID; overload;
    procedure Stop(RequestID: TRequestID);
    procedure StopAll;
    function Playing: Boolean;
    function WaitFor(RequestID: TRequestID): Boolean;
    function IsValid(RequestID: TRequestID): Boolean;
    function OrientationOf(Kind: TOrientedAction; Org, Trg: TPoint): TActorAction;
    property Actor: TAssistantActor read FActor;
    property Animator: TAssistantEngine read FAnimator;
    property IdleLevel: TIdleLevel read FIdleLevel write SetIdleLevel;
  published
    property AutoIdle: Boolean read FAutoIdle write SetAutoIdle default True;
    property Enabled: Boolean read FEnabled write SetEnabled default True;
    property OnPlayStart: TPlayEvent read FOnPlayStart write FOnPlayStart;
    property OnPlayStop: TPlayEvent read FOnPlayStop write FOnPlayStop;
    property OnIdle: TIdleEvent read FOnIdle write FOnIdle;
    property OnActorChange: TNotifyEvent read FOnActorChange write FOnActorChange;
  end;

  TRequestQueue = class(TList)
  private
    LastID: TRequestID;
    function NewRequest(AnimationIndex: Word): TRequestID;
    procedure RemoveRequest(RequestID: TRequestID);
    function RequestExists(RequestID: TRequestID): Boolean;
    function FetchCurrentTurn: TRequestID;
  end;

implementation

uses
  Forms;

{ TRequestQueue }

function TRequestQueue.NewRequest(AnimationIndex: Word): TRequestID;
begin
  Inc(LastID);
  Result := (LoWord(LastID) shl 16) or AnimationIndex;
  if Add(Pointer(Result)) < 0 then Result := 0;
end;

procedure TRequestQueue.RemoveRequest(RequestID: TRequestID);
var
  Index: Integer;
begin
  Index := IndexOf(Pointer(RequestID));
  if Index >= 0 then Delete(Index);
end;

function TRequestQueue.RequestExists(RequestID: TRequestID): Boolean;
begin
  Result := IndexOf(Pointer(RequestID)) >= 0;
end;

function TRequestQueue.FetchCurrentTurn: TRequestID;
begin
  Result := 0;
  if Count > 0 then
  begin
    Result := TRequestID(Items[0]);
    Delete(0);
  end;
end;

{ TAssistantDirector }

constructor TAssistantDirector.Create;
begin
  inherited Create;
  Randomize;
  FActor := TAssistantActor.Create;
  FActor.OnChange := ActorChanged;
  FAnimator := TAssistantEngine.Create;
  Animator.OnStop := AnimationStopped;
  IdleTimer := TTimer.Create(nil);
  IdleTimer.Enabled := False;
  IdleTimer.Interval := 3000;
  IdleTimer.OnTimer := IdleTimerExpire;
  Queue := TRequestQueue.Create;
  FAutoIdle := True;
  FEnabled := True;
  IdleLevel := Low(TIdleLevel);
  Returning := False;
  IdlingTime := False;
  CurAnim := nil;
  ID := 0;
end;

destructor TAssistantDirector.Destroy;
begin
  FAnimator.Terminate;
  if FAnimator.Suspended then
    FAnimator.Resume;
  FAnimator.WaitFor;
  IdleTimer.Free;
  FActor.Free;
  Queue.Free;
  inherited Destroy;
end;

procedure TAssistantDirector.SetEnabled(Value: Boolean);
begin
  if FEnabled <> Value then
  begin
    FEnabled := Value;
    if FEnabled then
      PlayQueuedRequest
    else
      IdleTimer.Enabled := False;
  end;
end;

procedure TAssistantDirector.SetAutoIdle(Value: Boolean);
begin
  if AutoIdle <> Value then
  begin
    FAutoIdle := Value;
    if not AutoIdle then
      IdleLevel := Low(TIdleLevel);
    GoToIdleState;
  end;
end;

procedure TAssistantDirector.SetIdleLevel(Value: TIdleLevel);
var
  M: Integer;
begin
  if (Value in [Low(TIdleLevel)..High(TIdleLevel)]) then
  begin
    FIdleLevel := Value;
    if (IdleLevel = Low(TIdleLevel)) and (ID = IdleID) and Animator.Active and
      Animator.HasInfiniteLoop then Stop(IdleID);
    M := Ord(High(TIdleLevel)) - Ord(FIdleLevel) + 1;
    IdleLevelTry := 4 * M + Random(10);
  end;
end;

procedure TAssistantDirector.IdleTimerExpire(Sender: TObject);
var
  Handled: Boolean;
  ThisLevel: Integer;
begin
  IdleTimer.Enabled := False;
  IdlingTime := True;
  try
    Handled := False;
    if Assigned(FOnIdle) then
    begin
      ThisLevel := Ord(FIdleLevel) - Ord(Low(TIdleLevel)) + 1;
      try
        FOnIdle(Self, ThisLevel, Handled);
      finally
        FIdleLevel := TIdleLevel(Ord(Low(TIdleLevel)) + ThisLevel - 1);
      end;
    end;
    if not Handled then
    begin
      IdleID := Queue.NewRequest(Actor.IndexOf(FIdleLevel));
      if IdleID > 0 then PlayQueuedRequest;
    end
    else if Queue.Count > 0 then
      IdleID := TRequestID(Queue.Items[Queue.Count-1])
    else
      IdleID := ID;
    Dec(IdleLevelTry);
    if IdleLevelTry = 0 then
      IdleLevel := Succ(IdleLevel);
    IdleTimer.Interval := 3000 + Random(1000) - Random(1000);
  finally
    IdlingTime := False;
    GoToIdleState;
  end;
end;

procedure TAssistantDirector.GoToIdleState;
begin
  IdleTimer.Enabled := FAutoIdle and FEnabled and not IdlingTime
    and Actor.Active and not Animator.Active and (Queue.Count = 0);
end;

procedure TAssistantDirector.ActorChanged(Sender: TObject);
begin
  StopAll;
  CurAnim := nil;
  if not Actor.Active then
  begin
    Animator.Clear;
    GoToIdleState;
  end;
  if Assigned(FOnActorChange) then
    FOnActorChange(Self);
end;

procedure TAssistantDirector.PlayQueuedRequest;
var
  NewAnim: PAnimationInfo;
begin
  if not (FEnabled and Actor.Active) or Animator.Active then Exit;
  IdleTimer.Enabled := False;
  NewAnim := nil;
  if (ID <> 0) and not Animator.UseExitBranching then
  begin
    if (not Animator.Done or (CurAnim.ExitType = etExitBranch)) and
      Animator.HasExitBranch then
    begin
      Returning := True;
      Animator.UseExitBranching := True;
      Animator.Active := True;
      Exit;
    end
    else if Animator.Done and (CurAnim.ExitType = etAnimation) then
    begin
      NewAnim := Actor.AnimationInfo[CurAnim^.ReturnIndex];
      Returning := (NewAnim <> nil);
    end;
  end;
  if not Returning then
  begin
    ID := Queue.FetchCurrentTurn;
    if ID <> 0 then NewAnim := Actor.AnimationInfo[LoWord(ID)];
  end;
  if Assigned(NewAnim) then
  begin
    if CurAnim <> NewAnim then
      Animator.LoadFromResourceID(Actor.Instance, NewAnim^.Resource)
    else
      Animator.CurrentFrame := 0;
    CurAnim := NewAnim;
    Animator.UseExitBranching := False;
    Animator.Active := True;
    if not Returning and (ID <> IdleID) and Assigned(FOnPlayStart) then
      FOnPlayStart(Self, ID);
  end
  else
    GoToIdleState;
end;

procedure TAssistantDirector.AnimationStopped(Sender: TObject);
var
  TriggerEvent: Boolean;
  StoppedID: TRequestID;
begin
  TriggerEvent := not Returning and (ID <> IdleID);
  Returning := False;
  StoppedID := ID;
  if ID = IdleID then
    IdleID := 0;
  try
    if TriggerEvent and Assigned(FOnPlayStop) then
      FOnPlayStop(Self, StoppedID);
  finally
    if Queue.Count > 0 then
      PlayQueuedRequest
    else
      GoToIdleState;
  end;
end;

function TAssistantDirector.Play(Index: Word): TRequestID;
begin
  Result := 0;
  if (Index > 0) and (Index <= Actor.AnimationCount) then
  begin
    IdleLevel := Low(TIdleLevel);
    if not Returning and (ID = IdleID) then
      Stop(IdleID);
    Result := Queue.NewRequest(Index);
    PlayQueuedRequest;
  end;
end;

function TAssistantDirector.Play(Name: String): TRequestID;
begin
  Result := Play(Actor.IndexOf(Name));
end;

function TAssistantDirector.Play(Action: TActorAction): TRequestID;
begin
  Result := Play(Actor.IndexOf(Action));
end;

function TAssistantDirector.WaitFor(RequestID: TRequestID): Boolean;
var
  OldID: TRequestID;
  DoneSteps: Byte;
begin
  Result := True;
  DoneSteps := 2;
  while Queue.RequestExists(RequestID) or ((ID = RequestID) and not Returning) do
  begin
    if Animator.HasInfiniteLoop then
    begin
      Result := False;
      Exit;
    end;
    OldID := ID;
    repeat
      Application.ProcessMessages;
    until not (Animator.Active and (OLDID = ID));
    if ID = RequestID then
    begin
      Dec(DoneSteps);
      if DoneSteps = 0 then Exit;
    end
  end;
end;

procedure TAssistantDirector.Stop(RequestID: TRequestID);
begin
  if Animator.Active then
  begin
    if (ID = RequestID) and not Animator.UseExitBranching then
    begin
      if Animator.Active then
      begin
        Animator.Active := False;
        Animator.CurrentFrame := Animator.CurrentFrame-1; 
      end;
    end
    else
      Queue.RemoveRequest(RequestID);
  end;
end;

procedure TAssistantDirector.StopAll;
begin
  if Animator.Active then
  begin
    Queue.Clear;
    if Animator.Active then
      Animator.Active := False;
  end;
  ID := 0;
end;

function TAssistantDirector.Playing: Boolean;
begin
  Result := Animator.Active or (Queue.Count > 0);
end;

function TAssistantDirector.IsValid(RequestID: TRequestID): Boolean;
begin
  Result := ((RequestID = ID) and Animator.Active and not Returning) or
            Queue.RequestExists(RequestID);
end;

function TAssistantDirector.OrientationOf(Kind: TOrientedAction;
  Org, Trg: TPoint): TActorAction;
type
  TDirection = (Down, Left, Right, Up);
const
  Actions: array[TOrientedAction, TDirection] of TActorAction = (
    (aaGestureDown, aaGestureLeft, aaGestureRight, aaGestureUp),
    (aaGlanceDown, aaGlanceLeft, aaGlanceRight, aaGlanceUp),
    (aaLookDown, aaLookLeft, aaLookRight, aaLookUp),
    (aaMoveDown, aaMoveLeft, aaMoveRight, aaMoveUp));
var
  Dir: TDirection;
begin
  if Abs(Org.x - Trg.x) > Abs(Org.y - Trg.y) then
    if Org.x > Trg.x then
      Dir := Right
    else
      Dir := Left
  else
    if Org.y > Trg.y then
      Dir := Up
    else
      Dir := Down;
  Result := Actions[Kind, Dir];
end;

end.
