unit oaEngine;

{ Office 97 Assistant Package v1.0 }
{ Copyright  1999 Kambiz R. Khojasteh }
{ email: Khojasteh@www.dci.co.ir }

interface

uses
  Windows, Classes, Graphics;

type

  TPaintFrameEvent = procedure(Sender: TObject; FrameBitmap: TBitmap) of object;

  TNormalBranch = record
    Precent: 0..100;
    Frame: Byte;
  end;

  TExitBranch = record
    Active: Boolean;
    Frame: Byte;
  end;

  TFrameItems = (fiImage, fiDelay, fiBranches, fiExitBranch, fiSound);
  TUpdatedItems = set of TFrameItems;

  TFrameData = class
  public
    UpdatedItems: TUpdatedItems;
    Image: TBitmap;
    Delay: Byte;
    Branches: array[1..4] of TNormalBranch;
    ExitBranch: TExitBranch;
    Sound: TMemoryStream;
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
  end;

  TAssistantEngine = class(TThread)
  private
    Bitmap, FrameBitmap: TBitmap;
    Delays: array[Byte] of Byte;
    Branches: array[Byte, 1..4] of TNormalBranch;
    ExitBranches: array[Byte] of TExitBranch;
    Sounds: array[Byte] of TMemoryStream;
    InternalChange: Boolean;
    ElapsedTime: DWord;
    VisibleFrame: Byte;
    FActive: Boolean;
    FFrameCount: Integer;
    FSpeed: Word;
    FDone: Boolean;
    FUseExitBranching: Boolean;
    FMute: Boolean;
    FCurrentFrame: Byte;
    FBgColor: TColor;
    FOnPaintFrame: TPaintFrameEvent;
    FOnChange: TNotifyEvent;
    FOnImageChange: TNotifyEvent;
    FOnStop: TNotifyEvent;
    function GetHasInfiniteLoop: Boolean;
    function GetHasSound: Boolean;
    function GetHasExitBranch: Boolean;
    function GetEmpty: Boolean;
    procedure SetBgColor(Value: TColor);
    function GetImageWidth: Word;
    function GetImageHeight: Word;
    procedure SetActive(Value: Boolean);
    procedure SetCurrentFrame(Value: Byte);
    procedure SetFrameData(Index: Byte; Value: TFrameData);
    function GetFrameData(Index: Byte): TFrameData;
    procedure RequestPaintFrame;
    procedure AnimationStopped;
    function FindNextFrame(Frame: Byte): Byte;
  protected
    procedure Execute; override;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    procedure LoadFromFile(const Filename: String);
    procedure SaveToFile(const Filename: String);
    procedure LoadFromResourceID(Instance: THandle; const ResID: Integer);
    procedure LoadFromResourceName(Instance: THandle; const ResName: String);
    procedure DeleteFrame(Index: Byte);
    function AddFrame(Frame: TFrameData): Byte;
    function InsertFrame(Index: Byte; Frame: TFrameData): Byte;
    function ValidateExitBranching: Integer;
    property Bitmaps: TBitmap read Bitmap;
    property FrameCount: Integer read FFrameCount;
    property ImageWidth: Word read GetImageWidth;
    property ImageHeight: Word read GetImageHeight;
    property Done: Boolean read FDone;
    property Empty: Boolean read GetEmpty;
    property BgColor: TColor read FBgColor write SetBgColor;
    property HasSound: Boolean read GetHasSound;
    property HasExitBranch: Boolean read GetHasExitBranch;
    property HasInfiniteLoop: Boolean read GetHasInfiniteLoop;
    property FrameData[Index: Byte]: TFrameData read GetFrameData write SetFrameData;
  published
    property Active: Boolean read FActive write SetActive default False;
    property Mute: Boolean read FMute write FMute default False;
    property Speed: Word read FSpeed write FSpeed default 100;
    property UseExitBranching: Boolean read FUseExitBranching write FUseExitBranching default False;
    property CurrentFrame: Byte read FCurrentFrame write SetCurrentFrame default 0;
    property OnPaintFrame: TPaintFrameEvent read FOnPaintFrame write FOnPaintFrame;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnImageChange: TNotifyEvent read FOnImageChange write FOnImageChange;
    property OnStop: TNotifyEvent read FOnStop write FOnStop;
  end;

function ScaleImageToRect(IR, R: TRect; Always: Boolean): TRect;
procedure DrawTiled(Canvas: TCanvas; Rect: TRect; G: TGraphic);

implementation

uses
  SysUtils, Messages, MMSystem, Forms, oaLZRWKH;

function ScaleImageToRect(IR, R: TRect; Always: Boolean): TRect;
var
  iW, iH: Integer;
  rW, rH: Integer;
begin
  iW := IR.Right - IR.Left;
  iH := IR.Bottom - IR.Top;
  rW := R.Right - R.Left;
  rH := R.Bottom - R.Top;
  if Always or (iW > rW) or (iH > rH) then
    if (rW / iW) < (rH / iH) then
    begin
      iH := MulDiv(iH, rW, iW);
      iW := MulDiv(iW, rW, iW);
    end
    else
    begin
      iW := MulDiv(iW, rH, iH);
      iH := MulDiv(iH, rH, iH);
    end;
  SetRect(Result, 0, 0, iW, iH);
  OffsetRect(Result, R.Left + (rW - iW) div 2, R.Top + (rH - iH) div 2);
end;

procedure DrawTiled(Canvas: TCanvas; Rect: TRect; G: TGraphic);
var
  R, Rows, C, Cols: Integer;
begin
  if (G <> nil) and (not G.Empty) then
  begin
    Rows := ((Rect.Bottom - Rect.Top) div G.Height) + 1;
    Cols := ((Rect.Right - Rect.Left) div G.Width) + 1;
    for R := 1 to Rows do
      for C := 1 to Cols do
        Canvas.Draw(Rect.Left + (C-1) * G.Width, Rect.Top + (R-1) * G.Height, G)
  end;
end;

{ TFrameData }

constructor TFrameData.Create;
begin
  inherited Create;
  UpdatedItems := [Low(TFrameItems)..High(TFrameItems)];
  Image := TBitmap.Create;
  Sound := TMemoryStream.Create;
  Delay := 10;
end;

destructor TFrameData.Destroy;
begin
  Image.Free;
  Sound.Free;
  inherited Destroy;
end;

procedure TFrameData.Clear;
begin
  UpdatedItems := [Low(TFrameItems)..High(TFrameItems)];
  Image.Free;
  Image := TBitmap.Create;
  Sound.Clear;
  Delay := 10;
  FillChar(Branches, SizeOf(Branches), 0);
  FillChar(ExitBranch, SizeOf(ExitBranch), 0);
end;

{ TAssistantEngine }

const
  AAIdentifier: LongInt =
    ((((((Ord('1') SHL 8)+Ord('K')) SHL 8)+Ord('R')) SHL 8)+Ord('K'));

constructor TAssistantEngine.Create;
begin
  Randomize;
  Bitmap := TBitmap.Create;
  FrameBitmap := TBitmap.Create;
  FActive := False;
  FSpeed := 100;
  FMute := False;
  FUseExitBranching := False;
  FFrameCount := 0;
  FCurrentFrame := 0;
  FBgColor := clBlack;
  VisibleFrame := 0;
  InternalChange := False;
  FreeOnTerminate := True;
  inherited Create(True);
end;

destructor TAssistantEngine.Destroy;
var
  I: Integer;
begin
  Bitmap.Free;
  FrameBitmap.Free;
  for I := 0 to FFrameCount-1 do
    if Assigned(Sounds[I]) then Sounds[I].Free;
  inherited Destroy;
end;

procedure TAssistantEngine.Clear;
var
  I: Integer;
begin
  Active := False;
  Bitmap.Free;
  Bitmap := TBitmap.Create;
  for I := 0 to FFrameCount-1 do
    if Assigned(Sounds[I]) then Sounds[I].Free;
  FillChar(Delays, SizeOf(Delays), 10);
  FillChar(Branches, SizeOf(Branches), 0);
  FillChar(Sounds, SizeOf(Delays), 0);
  FillChar(ExitBranches, SizeOf(ExitBranches), -1);
  FFrameCount := 0;
  FCurrentFrame := 0;
  FBgColor := clBlack;
  FrameBitmap.Free;
  FrameBitmap := TBitmap.Create;
  FDone := False;
  VisibleFrame := 0;
  if not InternalChange and not Terminated and Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TAssistantEngine.LoadFromStream(Stream: TStream);
const
  ReadErrorMsg = 'Invalid animation data';
var
  I: Integer;
  TmpStream: TMemoryStream;
  Identifier, SoundSize: LongInt;
begin
  Stream.Position := 0;
  Stream.Read(Identifier, SizeOf(Identifier));
  if Identifier <> AAIdentifier then
    raise Exception.Create(ReadErrorMsg);
  TmpStream := TMemoryStream.Create;
  InternalChange := True;
  Clear;
  try
    DecompressStream(Stream, TmpStream);
    TmpStream.Position := 0;
    TmpStream.Read(FFrameCount, SizeOf(FFrameCount));
    TmpStream.Read(Delays, FFrameCount * SizeOf(Delays[0]));
    TmpStream.Read(Branches, 4 * FFrameCount * SizeOf(Branches[0,1]));
    TmpStream.Read(ExitBranches, FFrameCount * SizeOf(ExitBranches[0]));
    TmpStream.Read(FBgColor, SizeOf(FBgColor));
    for I := 0 to FFrameCount-1 do
    begin
      TmpStream.Read(SoundSize, SizeOf(SoundSize));
      if SoundSize > 0 then
      begin
        Sounds[I] := TMemoryStream.Create;
        Sounds[I].CopyFrom(TmpStream, SoundSize);
      end;
    end;
    try
      Bitmap.LoadFromStream(TmpStream);
    except
      raise Exception.Create(ReadErrorMsg);
    end;
    FrameBitmap.Width := Bitmap.Width div FFrameCount;
    FrameBitmap.Height := Bitmap.Height;
    FrameBitmap.Transparent := True;
    FrameBitmap.TransparentColor := FBgColor;
    if Assigned(FOnChange) then FOnChange(Self);
  finally
    TmpStream.Free;
    InternalChange := False;
  end;
end;

procedure TAssistantEngine.SaveToStream(Stream: TStream);
var
  I: Integer;
  TmpStream: TMemoryStream;
  Identifier, SoundSize: LongInt;
begin
  if (FFrameCount = 0) or Bitmap.Empty then Exit;
  TmpStream := TMemoryStream.Create;
  try
    TmpStream.Write(FFrameCount, SizeOf(FFrameCount));
    TmpStream.Write(Delays, FFrameCount * SizeOf(Delays[0]));
    TmpStream.Write(Branches, 4 * FFrameCount * SizeOf(Branches[0,1]));
    TmpStream.Write(ExitBranches, FFrameCount * SizeOf(ExitBranches[0]));
    TmpStream.Write(FBgColor, SizeOf(FBgColor));
    for I := 0 to FFrameCount-1 do
    begin
      if Assigned(Sounds[I]) then
        SoundSize := Sounds[I].Size
      else
        SoundSize := 0;
      TmpStream.Write(SoundSize, SizeOf(SoundSize));
      if SoundSize > 0 then
      begin
        Sounds[I].Position := 0;
        TmpStream.CopyFrom(Sounds[I], SoundSize);
      end;
    end;
    Bitmap.SaveToStream(TmpStream);
    TmpStream.Position := 0;
    Identifier := AAIdentifier;
    Stream.Write(Identifier, SizeOf(Identifier));
    CompressStream(TmpStream, Stream);
  finally
    TmpStream.Free;
  end;
end;

procedure TAssistantEngine.LoadFromFile(const Filename: String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(Filename, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TAssistantEngine.SaveToFile(const Filename: String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(Filename, fmCreate or fmShareExclusive);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TAssistantEngine.LoadFromResourceID(Instance: THandle; const ResID: Integer);
var
  Stream: TResourceStream;
begin
  Stream := TResourceStream.CreateFromID(Instance, ResID, RT_RCDATA);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TAssistantEngine.LoadFromResourceName(Instance: THandle; const ResName: String);
var
  Stream: TResourceStream;
begin
  Stream := TResourceStream.Create(Instance, ResName, RT_RCDATA);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TAssistantEngine.DeleteFrame(Index: Byte);
var
  SrcRect, DstRect: TRect;
  I, J, W, H: Integer;
begin
  if not FActive and (Index < FFrameCount) then
  begin
    if FFrameCount = 1 then
      Clear
    else
    begin
      W := ImageWidth;
      H := ImageHeight;
      SetRect(DstRect, Index * W, 0, Bitmap.Width - W, H);
      SetRect(SrcRect, (Index+1) * W, 0, Bitmap.Width, H);
      if Index < (FFrameCount-1) then
        Bitmap.Canvas.CopyRect(DstRect, Bitmap.Canvas, SrcRect);
      Bitmap.Width := Bitmap.Width - W;
      if Assigned(Sounds[Index]) then
        Sounds[Index].Free;
      for I := Index+1 to FFrameCount-1 do
      begin
        Delays[I-1] := Delays[I];
        Sounds[I-1] := Sounds[I];
        for J := 1 to 4 do
        begin
          Branches[I-1,J] := Branches[I,J];
          if Branches[I-1,J].Precent > 0 then
            if Branches[I-1,J].Frame > Index then
              Dec(Branches[I-1,J].Frame)
            else if Branches[I-1,J].Frame = Index then
              Branches[I-1,J].Precent := 0;
        end;
        ExitBranches[I-1] := ExitBranches[I];
        if ExitBranches[I-1].Active then
          if ExitBranches[I-1].Frame > Index then
            Dec(ExitBranches[I-1].Frame)
          else if ExitBranches[I-1].Frame = Index then
            ExitBranches[I-1].Active := False;
      end;
      Sounds[FFrameCount-1] := nil;
      Dec(FFrameCount);
      if FCurrentFrame >= FFrameCount then
        FCurrentFrame := FFrameCount-1;
      if Assigned(FOnChange) then
        FOnChange(Self);
    end;
  end;
end;

function TAssistantEngine.AddFrame(Frame: TFrameData): Byte;
begin
  Result := InsertFrame(FFrameCount, Frame);
end;

function TAssistantEngine.InsertFrame(Index: Byte; Frame: TFrameData): Byte;
var
  SrcRect, DstRect: TRect;
  I, J, W, H: Integer;
begin
  if not FActive and (FFrameCount < High(Delays)) then
  begin
    if Index > FFrameCount then
      Index := FFrameCount;
    W := ImageWidth;
    H := ImageHeight;
    if FFrameCount = 0 then
    begin
      Bitmap.Assign(Frame.Image);
      Bitmap.Transparent := False;
      FrameBitmap.Width := Bitmap.Width;
      FrameBitmap.Height := Bitmap.Height;
      FrameBitmap.Transparent := True;
      FrameBitmap.TransparentColor := FBgColor;
    end
    else
    begin
      SetRect(DstRect, (Index+1) * W, 0, Bitmap.Width + W, H);
      SetRect(SrcRect, Index * W, 0, Bitmap.Width, H);
      Bitmap.Width := Bitmap.Width + W;
      if Index < FFrameCount then
        Bitmap.Canvas.CopyRect(DstRect, Bitmap.Canvas, SrcRect);
      SetRect(SrcRect, Index * W, 0, (Index+1) * W, H);
      for I := FFrameCount-1 downto Index do
      begin
        Delays[I+1] := Delays[I];
        Sounds[I+1] := Sounds[I];
        for J := 1 to 4 do
        begin
          Branches[I+1,J] := Branches[I,J];
          if (Branches[I+1,J].Precent > 0) and (Branches[I+1,J].Frame >= Index) then
            Inc(Branches[I+1,J].Frame);
        end;
        ExitBranches[I+1] := ExitBranches[I];
        if ExitBranches[I+1].Active and (ExitBranches[I+1].Frame >= Index) then
          Inc(ExitBranches[I+1].Frame);
      end;
      Sounds[Index] := nil;
    end;
    Inc(FFrameCount);
    FrameData[Index] := Frame;
    Result := Index;
  end
  else
    Result := High(Delays);
end;

procedure TAssistantEngine.SetFrameData(Index: Byte; Value: TFrameData);
var
  SrcRect, DstRect: TRect;
  OldTransparent: Boolean;
  I: Integer;
begin
  if not FActive and (Index < FFrameCount) then
  begin
    if fiImage in Value.UpdatedItems then
    begin
      SetRect(SrcRect, 0, 0, Value.Image.Width, Value.Image.Height);
      SetRect(DstRect, 0, 0, ImageWidth, ImageHeight);
      OffsetRect(DstRect, Index * ImageWidth, 0);
      Bitmap.Canvas.Brush.Color := FBgColor;
      Bitmap.Canvas.FillRect(DstRect);
      DstRect := ScaleImageToRect(SrcRect, DstRect, False);
      OldTransparent := Value.Image.Transparent;
      Bitmap.Canvas.StretchDraw(DstRect, Value.Image);
      Value.Image.Transparent := OldTransparent;
    end;
    if fiDelay in Value.UpdatedItems then
      Delays[Index] := Value.Delay;
    if fiBranches in Value.UpdatedItems then
      for I := 1 to 4 do
        Branches[Index, I] := Value.Branches[I];
    if fiExitBranch in Value.UpdatedItems then
      ExitBranches[Index] := Value.ExitBranch;
    if fiSound in Value.UpdatedItems then
    begin
      if Assigned(Sounds[Index]) then
        Sounds[Index].Free;
      Sounds[Index] := nil;
      if Assigned(Value.Sound) and (Value.Sound.Size > 0) then
      begin
        Sounds[Index] := TMemoryStream.Create;
        Sounds[Index].CopyFrom(Value.Sound, 0);
      end;
    end;
    if (Value.UpdatedItems <> []) and Assigned(FOnChange) then
      FOnChange(Self);
    if (fiImage in Value.UpdatedItems) and Assigned(FOnImageChange) then
      FOnImageChange(Self);
  end;
end;

function TAssistantEngine.GetFrameData(Index: Byte): TFrameData;
var
  SrcRect, DstRect: TRect;
  I: Integer;
begin
  Result := TFrameData.Create;
  if Index < FFrameCount then
  begin
    Result.Image.Width := ImageWidth;
    Result.Image.Height := ImageHeight;
    SetRect(DstRect, 0, 0, Result.Image.Width, Result.Image.Height);
    SrcRect := DstRect;
    OffsetRect(SrcRect, Index * Result.Image.Width, 0);
    Result.Image.Canvas.CopyRect(DstRect, Bitmap.Canvas, SrcRect);
    Result.Delay := Delays[Index];
    for I := 1 to 4 do
      Result.Branches[I] := Branches[Index, I];
    Result.ExitBranch := ExitBranches[Index];
    if Assigned(Sounds[Index]) then
      Result.Sound.CopyFrom(Sounds[Index], 0);
  end;
end;

function TAssistantEngine.GetHasInfiniteLoop: Boolean;
var
  I, J: Integer;
begin
  Result := False;
  for I := 0 to FFrameCount-1 do
    for j := 1 to 4 do
      if (Branches[I,J].Precent = 100) and (Branches[I,J].Frame <= I) then
      begin
        Result := True;
        Exit;
      end;
end;

function TAssistantEngine.GetHasSound: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to FFrameCount-1 do
    if Assigned(Sounds[I]) and (Sounds[I].Size > 0) then
    begin
      Result := True;
      Exit;
    end;
end;

function TAssistantEngine.GetHasExitBranch: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to FFrameCount-1 do
    if ExitBranches[I].Active then
    begin
      Result := True;
      Exit;
    end;
end;

function TAssistantEngine.GetEmpty: Boolean;
begin
  Result := (FFrameCount = 0) or Bitmap.Empty;
end;

procedure TAssistantEngine.SetBgColor(Value: TColor);
begin
  if FBgColor <> Value then
  begin
    FBgColor := Value;
    FrameBitmap.Transparent := True;
    FrameBitmap.TransparentColor := Value;
    if Assigned(FOnChange) then
      FOnChange(Self);
    if Assigned(FOnImageChange) then
      FOnImageChange(Self);
  end;
end;

function TAssistantEngine.GetImageWidth: Word;
begin
  if (FFrameCount = 0) or Bitmap.Empty then
    Result := 0
  else
    Result := Bitmap.Width div FFrameCount;
end;

function TAssistantEngine.GetImageHeight: Word;
begin
  if (FFrameCount = 0) or Bitmap.Empty then
    Result := 0
  else
    Result := Bitmap.Height;
end;

procedure TAssistantEngine.SetActive(Value: Boolean);
var
  Msg: TMsg;
begin
  if FActive <> Value then
  begin
    FActive := Value;
    if FActive then
    begin
      FDone := False;
      if Suspended then Resume;
    end
    else
    begin
      sndPlaySound(nil, SND_MEMORY);
      repeat
        PeekMessage(Msg, Application.Handle, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE);
        PeekMessage(Msg, Application.Handle, WM_MOUSEFIRST, WM_MOUSELAST, PM_REMOVE);
      until (Terminated or Suspended or FActive);
    end;
  end;
end;

procedure TAssistantEngine.SetCurrentFrame(Value: Byte);
begin
  if Value >= FFrameCount then
    Value := FFrameCount-1;
  if (FCurrentFrame <> Value) and (Value < FFrameCount) then
  begin
    FCurrentFrame := Value;
    Synchronize(RequestPaintFrame);
  end;
end;

function TAssistantEngine.ValidateExitBranching: Integer;
var
  B: Byte;
  F, BF: Integer;
begin
  Result := -1;
  for F := 0 to FFrameCount-2 do
  begin
    if ExitBranches[F].Active then
    begin
      BF := ExitBranches[F].Frame;
      while BF <= F do
      begin
        if not ExitBranches[BF].Active or (ExitBranches[BF].Frame in [BF..F]) then
        begin
          Result := BF;
          Exit;
        end
        else
          BF := ExitBranches[BF].Frame;
      end;
    end
    else
    begin
      for B := 1 to 4 do
        if (Branches[F,B].Precent > 0) and (Branches[F,B].Frame <= F) then
        begin
          Result := F;
          Exit;
        end;
    end;
  end;
end;

procedure TAssistantEngine.AnimationStopped;
begin
  if not Terminated and Assigned(FOnStop) then
    FOnStop(Self);
end;

procedure TAssistantEngine.RequestPaintFrame;
var
  SrcRect, DstRect: TRect;
begin
  ElapsedTime := GetTickCount;
  try
    SetRect(DstRect, 0, 0, FrameBitmap.Width, FrameBitmap.Height);
    SrcRect := DstRect;
    OffsetRect(SrcRect, FCurrentFrame * FrameBitmap.Width, 0);
    FrameBitmap.Canvas.CopyRect(DstRect, Bitmap.Canvas, SrcRect);
    if not Terminated and Assigned(FOnPaintFrame) then
      FOnPaintFrame(Self, FrameBitmap);
    if not Terminated and Assigned(Sounds[FCurrentFrame]) and FActive and not FMute then
      sndPlaySound(Sounds[FCurrentFrame].Memory, SND_MEMORY or SND_NOSTOP or SND_ASYNC);
  finally
    ElapsedTime := (GetTickCount - ElapsedTime) div 10;
  end;
end;

function TAssistantEngine.FindNextFrame(Frame: Byte): Byte;
var
  B, P: Byte;
  Branched: Boolean;
begin
  Result := Frame;
  if FUseExitBranching and ExitBranches[Frame].Active then
    if Frame < FFrameCount-1 then
      Result := ExitBranches[Frame].Frame
    else
      FDone := True
  else
  begin
    P := 100;
    Branched := False;
    for B := 1 to 4 do
      if Branches[Frame,B].Precent = 0 then
        Break
      else if Random(P) < Branches[Frame,B].Precent then
      begin
        Result := Branches[Frame,B].Frame;
        Branched := True;
        Break;
      end
      else
        Dec(P, Branches[Frame,B].Precent);
    if not Branched then
    begin
      if Frame < FFrameCount-1 then
        Result := Frame + 1
      else
        FDone := True;
    end;
  end;
end;

procedure TAssistantEngine.Execute;
begin
  while not Terminated do
    if FActive and (FSpeed > 0) then
    begin
      if (Delays[FCurrentFrame] > 0) and
         (not UseExitBranching or (VisibleFrame <> CurrentFrame)) then
      begin
        Synchronize(RequestPaintFrame);
        VisibleFrame := FCurrentFrame;
        if Delays[FCurrentFrame] > ElapsedTime then
          Sleep(MulDiv(Delays[FCurrentFrame] - ElapsedTime, 1000, FSpeed));
      end;
      FCurrentFrame := FindNextFrame(FCurrentFrame);
      if FDone then
      begin
        FActive := False;
        if Delays[FCurrentFrame] = 0 then
          FCurrentFrame := VisibleFrame;
      end;
    end
    else if not (FActive or Terminated) then
    begin
      Synchronize(AnimationStopped);
      if not FActive then Suspend;
    end;
end;

end.
